#include <X11/Intrinsic.h>
#include "global.h"
#include <errno.h>
#include "map.h"

PUBLIC(void, MainLoop, (void));
PUBLIC(void, HandleButtonPress, (XEvent *event));
PUBLIC(void, HandleKeyPress, (XEvent *event));
PUBLIC(void, HandleExpose, (void));
PUBLIC(void, HandleClientMessage, (XEvent *event, Bool *cont_flag));
PUBLIC(void, HandleVisibility, (XEvent *event));

EXTERN(void, remap, (struct map *map));
EXTERN(Bool, RegrabKey, (void));
EXTERN(void, PutText, (char *text, int len));
EXTERN(void, LaunchXLock, (void));
EXTERN(void, Quit, (int status));

extern Display *dpy;
extern char *dpyname, *app_name;
extern Window win;
extern KeyCode keycode, quit_keycode, xlock_keycode;
extern struct map *current_map;
extern Bool grabbed, popup;
extern Atom wmproto, protocol[2];

#ifdef HAVE_XPM
extern Window icon_win;
#endif

void MainLoop()
{
 XEvent event;
 Bool cont_flag;

 for (cont_flag = True; cont_flag; ) {
	XNextEvent(dpy, &event);
	switch (event.type) {
		case ButtonPress :
			HandleButtonPress(&event);
		break;
		case KeyPress :
			HandleKeyPress(&event);
		break;
		case Expose :
			if (!event.xexpose.count)
				HandleExpose();
		break;
		case VisibilityNotify :
			if (popup)
				HandleVisibility(&event);
		break;
		case ClientMessage :
			HandleClientMessage(&event, &cont_flag);
		break;
	}
 }

 return;
}

void HandleButtonPress(event)
	XEvent *event;
{
 struct map *new_map;

 switch (event->xbutton.button) {
	case 1:
		for (new_map = current_map->next; !new_map->enabled;
		     new_map = new_map->next);
		remap(new_map);
	break;
	case 2:
		grabbed = RegrabKey();
	break;
	case 3:
		for (new_map = current_map->prev; !new_map->enabled;
		     new_map = new_map->prev);
		remap(new_map);
	break;
 }

 return;
}

void HandleKeyPress(event)
	XEvent *event;
{
 struct map *new_map;

 if (event->xkey.keycode == quit_keycode &&
     (event->xkey.window == win
#ifdef HAVE_XPM
				|| event->xkey.window == icon_win
#endif
								  ))
		Quit(0);

 if (event->xkey.keycode == xlock_keycode &&
     (event->xkey.state & ControlMask))
	LaunchXLock();

 if (event->xkey.keycode == keycode) {
	for (new_map = current_map->next; !new_map->enabled;
	     new_map = new_map->next);
	remap(new_map);
 }
 
 return;
}

void HandleExpose()
{
 PutText(current_map->text, current_map->len);

 return;
}

void HandleClientMessage(event, cont_flag)
	XEvent *event;
	Bool *cont_flag;
{

 if (event->xclient.message_type == wmproto &&
     (event->xclient.data.l[0] == protocol[0] ||
      event->xclient.data.l[0] == protocol[1])) {
	struct map *ptr;

	for (ptr = current_map; ptr->index != STANDARD_MAP; ptr = ptr->next);
	remap(ptr);
	*cont_flag = False;
 }

 return;
}

void HandleVisibility(event)
	XEvent *event;
{
 if (event->xvisibility.state != VisibilityUnobscured)
	XRaiseWindow(dpy, win);

 return;
}

#ifndef HAVE_STRERROR

const char *__errlist[] = {
	"error 0",
	"not owner",
	"no such file or directory",
	"no such process",
	"interrupted system call",
	"I/O error",
	"no such device or address",
	"argument list too long",
	"exec format error",
	"bad file number",
	"no child process",
	"no more processes",
	"not enough memory",
	"permission denied",
	"bad address",
	"block device required",
	"mount devices busy",
	"file exists",
	"cross-device link",
	"no such device",
	"not a directory",
	"is a directory",
	"invalid argument",
	"file table overflow",
	"too many open files",
	"not a typewriter",
	"text file busy",
	"file too big",
	"no space left on device",
	"illegal seek",
	"read-only file system",
	"too many links",
	"broken pipe",
	"math argument",
	"result too large",
	"I/O stream empty",
	"vax/vms specific error code nontranslatable error"
};

int __nerr = sizeof(__errlist) / sizeof(__errlist[0]);

const char *strerror(errno)
	int errno;
{
 if (errno >= 0 && errno < __nerr)
	return sys_errlist[errno];

 return "Unknown UNIX error";
}

#endif

int IOHandler(display)
	Display *display;
{
 warn("Fatal I/O Error %d (%s). Exiting ..", errno, strerror(errno));
 XCloseDisplay(display);

 if ((dpy = XOpenDisplay(dpyname)))
	Quit(255);

 exit(1);
}
