#ifdef HAVE_STRINGS_H
#include <strings.h>
#endif
#include <stdlib.h>
#include <X11/Intrinsic.h>

#ifdef HAVE_XPM
#include <xpm.h>
#else
#include "mask.xbm"
#endif

#include "global.h"
#include "map.h"

#ifndef HAVE_XPM
Pixmap icon_mask;
#endif

static char *location;
extern int icon_size;

extern Display *dpy;
extern Window win;
extern Colormap cmap;
extern Bool nokoi, noiso, noalt, nowin;
extern struct map *current_map;

#ifdef HAVE_XPM
extern Window icon_win;
#endif

extern char *app_name;

#ifndef MAXPATHLEN
#define MAXPATHLEN	1024
#endif

void GetWMIconSize()
{
 XIconSize *size;
 int i, count;

 if (XGetIconSizes(dpy, DefaultRootWindow(dpy), &size, &count)) {
	for (i=0; i<count; i++)
		if (size[i].max_width > 63 && size[i].max_height > 63) {
			icon_size = 64;
			XFree(size);
			return;
		}
	icon_size = 48;
	XFree(size);
 } else
	icon_size = 64;

 return;
}

Bool CreatePixmap(name, homedir, image)
	char *name;
	char *homedir;
	Pixmap *image;
{
 char path[MAXPATHLEN];

#ifdef HAVE_XPM

 XpmAttributes xpm_attr;

 bzero(&xpm_attr, sizeof(XpmAttributes));
 xpm_attr.valuemask = XpmColormap | XpmCloseness;
 xpm_attr.colormap = cmap;
 xpm_attr.closeness = 65535;

#else

 int r;

#endif

 if (homedir) {

#ifdef HAVE_XPM

	sprintf(path, "%s/.xes/%s/%s.xpm", homedir, location, name); 

	if (XpmReadFileToPixmap(dpy, icon_win, path,
				image, NULL, &xpm_attr) == XpmSuccess)
#else

	sprintf(path, "%s/.xes/%s/%s.xbm", homedir, location, name); 

	if (XReadBitmapFile(dpy, win, path,
			    (unsigned int*) &r, (unsigned int*) &r,
			    image, &r, &r) == BitmapSuccess)
#endif
		return True;
 }

#ifdef HAVE_XPM

 sprintf(path, "%s/%s/%s.xpm", XESLIBDIR, location, name);

	if (XpmReadFileToPixmap(dpy, icon_win, path,
				image, NULL, &xpm_attr) == XpmSuccess)
#else

 sprintf(path, "%s/%s/%s.xbm", XESLIBDIR, location, name);

	if (XReadBitmapFile(dpy, win, path,
			    (unsigned int*) &r, (unsigned int*) &r,
			    image, &r, &r) == BitmapSuccess)
#endif
		return True;

 return False;
}

Bool create_icons()
{
 Pixmap p;
 char *home;

 if (!icon_size)
	GetWMIconSize();

#ifdef HAVE_XPM

 if (!(icon_win = XCreateSimpleWindow(dpy, DefaultRootWindow(dpy),
				      0, 0, icon_size,icon_size, 0,
				      CopyFromParent, CopyFromParent))) {
	warn("can't create icon window");
	return False;
 }

#endif

 location = (icon_size == 64) ? "64x64" : "48x48";
 home = getenv("HOME");

 if (!CreatePixmap("standard", home, &p)) {
	warn("can't create icon pixmap for STANDARD encoding");
	return False;
 }
 current_map->pixmap = p;

 if (!nokoi) {
	if (!CreatePixmap("koi", home, &p)) {
		warn("can't create icon pixmap for KOI encoding");
		XFreePixmap(dpy, current_map->pixmap);
		return False;
	}
	current_map->next->pixmap = p;
 }

 if (!noiso) {
	if (!CreatePixmap("iso", home, &p)) {
		warn("can't create icon pixmap for ISO encoding");
		XFreePixmap(dpy, current_map->pixmap);
		if (!nokoi)
			XFreePixmap(dpy, current_map->next->pixmap);
		return False;
	}
	current_map->next->next->pixmap = p;
 }

 if (!noalt) {
	if (!CreatePixmap("alt", home, &p)) {
		warn("can't create icon pixmap for ALT encoding");
		XFreePixmap(dpy, current_map->pixmap);
		if (!nokoi)
			XFreePixmap(dpy, current_map->next->pixmap);
		if (!noiso)
			XFreePixmap(dpy, current_map->next->next->pixmap);
		return False;
	}
	current_map->prev->prev->pixmap = p;
 }

 if (!nowin) {
	if (!CreatePixmap("win", home, &p)) {
		warn("can't create icon pixmap for WIN encoding");
		XFreePixmap(dpy, current_map->pixmap);
		if (!nokoi)
			XFreePixmap(dpy, current_map->next->pixmap);
		if (!noiso)
			XFreePixmap(dpy, current_map->next->next->pixmap);
		if (!noalt)
			XFreePixmap(dpy, current_map->prev->prev->pixmap);
		return False;
	}
	current_map->prev->pixmap = p;
 }

#ifndef HAVE_XPM

 if ((icon_mask = XCreateBitmapFromData(dpy, win, (char*) mask_bits,
					mask_width, mask_height)) == None) {
	warn("can't create pixmap mask for icon");
	XFreePixmap(dpy, current_map->pixmap);
	if (!nokoi)
		XFreePixmap(dpy, current_map->next->pixmap);
	if (!noiso)
		XFreePixmap(dpy, current_map->next->next->pixmap);
	if (!noalt)
		XFreePixmap(dpy, current_map->prev->prev->pixmap);
	if (!nowin)
		XFreePixmap(dpy, current_map->prev->pixmap);

	return False;
 }

#else

 XSetWindowColormap(dpy, icon_win, cmap);

#endif

 return True;
}

void set_icon_name(name)
	char *name;
{
 char buf[255];

 (void) sprintf(buf, "%s -- %s", app_class, name);
 XStoreName(dpy, win, buf);

 return;
}
