#ifdef HAVE_STRINGS_H
#include <strings.h>
#endif
#include <stdlib.h>

#include <X11/Intrinsic.h>
#include "global.h"
#include "key.h"
#include "keys.h"
#include "koi8.h"
#include "iso8859.h"
#include "alt.h"
#include "win.h"
#include "map.h"

PUBLIC(KeySym*, first_map, (int map_id));
EXTERN(void, Quit, (int status));

struct map *current_map;
int first, max, key_per_code;

extern Display *dpy;
extern char *app_name;
extern Bool nokoi, noiso, noalt, nowin;

void store_maps()
{
 struct map *new, *last;
 XDisplayKeycodes(dpy, &first, &max);

 if ((new = (struct map*) malloc(sizeof(struct map))) == NULL) {
	warn("Out of memory");
	Quit(1);
 }
 current_map = last = new;
 new->index = STANDARD_MAP;
 new->enabled = True;
 new->map = XGetKeyboardMapping(dpy, first, max-first+1, &key_per_code);
 new->text = "Standard";
 new->len = 8;

 if ((new = (struct map*) malloc(sizeof(struct map))) == NULL) {
	warn("Out of memory");
	Quit(1);
 }
 if ((new->enabled = !nokoi)) {
	new->map = first_map(KOI8_MAP);
	new->text = "koi8 (rus)";
	new->len = 10;
 }
 new->index = KOI8_MAP;
 new->prev = last;
 last->next = new;
 last = new;

 if ((new = (struct map*) malloc(sizeof(struct map))) == NULL) {
	warn("Out of memory");
	Quit(1);
 }
 if ((new->enabled = !noiso)) {
	new->map = first_map(ISO8859_MAP);
	new->text = "iso8859 (rus)";
	new->len = 13;
 }
 new->index = ISO8859_MAP;
 new->prev = last;
 last->next = new;
 last = new;

 if ((new = (struct map*) malloc(sizeof(struct map))) == NULL) {
	warn("Out of memory");
	Quit(1);
 }
 if ((new->enabled = !noalt)) {
	new->map = first_map(ALT_MAP);
	new->text = "alt (rus)";
	new->len = 9;
 }
 new->index = ALT_MAP;
 new->prev = last;
 last->next = new;
 last = new;

 if ((new = (struct map*) malloc(sizeof(struct map))) == NULL) {
	warn("Out of memory");
	Quit(1);
 }
 if ((new->enabled = !nowin)) {
	new->map = first_map(WIN_MAP);
	new->text = "win (rus)";
	new->len = 9;
 }
 new->index = WIN_MAP;
 new->prev = last;
 new->next = current_map;
 current_map->prev = last->next = new;

 return;
}

KeySym *first_map(map_id)
	int map_id;
{
 KeySym *syms, *standard_map;
 Map *map;
 int i, j;

#define MAP_SIZE	((max-first+1)*key_per_code*sizeof(KeySym))

 if ((syms = (KeySym*) malloc(MAP_SIZE)) == NULL) {
	warn("Out of memory");
	Quit(1);
 }

 standard_map = current_map->map;

#ifdef HAVE_BCOPY
 bcopy(standard_map, syms, MAP_SIZE);
#else
 memcpy(syms, standard_map, MAP_SIZE);
#endif

#undef MAP_SIZE

 switch (map_id) {
	case KOI8_MAP :
		map = koi8;
	break;
	case ISO8859_MAP :
		map = iso8859;
	break;
	case ALT_MAP :
		map = alt;
	break;
	case WIN_MAP :
		map = ___win;
	break;
	default :
		return NULL;
 }

#define IDX	((j-first)*key_per_code)

 for (i=0; i<20; i++)
	for (j=first; j<max-first+1; j++) {
		if (standard_map[IDX] == standard[i].sym[0])
			syms[IDX] = map[i].sym[0]; else
		if (standard_map[IDX] == standard[i].sym[1])
			syms[IDX] = map[i].sym[1];
		if (standard_map[IDX+1] == standard[i].sym[0])
			syms[IDX+1] = map[i].sym[0]; else
		if (standard_map[IDX+1] == standard[i].sym[1])
			syms[IDX+1] = map[i].sym[1];
 }

 for (i=20; i<MAX_ENTRY; i++)
	for (j=first; j<max-first+1; j++)
		if (standard_map[IDX] == standard[i].sym[0] ||
		    standard_map[IDX] == standard[i].sym[1]) {
			syms[IDX] = map[i].sym[0];
			syms[IDX+1] = map[i].sym[1];
		}

 return syms;
}
