#include <stdlib.h>
#include <X11/Intrinsic.h>
#include <X11/keysym.h>
#include <X11/cursorfont.h>
#include "global.h"
#include "map.h"

PUBLIC(void, InitKeycodes, (void));
PUBLIC(void, remap, (struct map *map));
PUBLIC(void, PutText, (char *text, int len));
PUBLIC(Bool, GrabKey, (void));
PUBLIC(Bool, RegrabKey, (void));
PUBLIC(void, ChangeIcon, (Pixmap pixmap));
PUBLIC(Bool, get_x_y, (char *res, char *X_pos, char *Y_pos,
		       int X_def, int Y_def, int *X, int *Y));
PUBLIC(void, LaunchXLock, (void));

EXTERN(char*, GetResource, (char *name));
EXTERN(void, set_icon_name, (char *name));

KeyCode keycode, caps_lock, num_lock, scroll_lock,
	quit_keycode, xlock_keycode;

extern Display *dpy;
extern int scr, first, max, key_per_code, win_w, win_h, y_pos;
extern struct map *current_map;
extern Colormap cmap;
extern char *app_name, *keyname, *xlock_string;
extern Bool have_icons, popup, grabbed, fvwm95;
extern Window win;

#ifdef HAVE_XPM
extern Window icon_win;
#endif

extern XWMHints h;
extern GC gc;
extern Font font;
extern XFontStruct *fs;

extern Atom protocol, wmproto;

void InitKeycodes()
{
 caps_lock = XKeysymToKeycode(dpy, XK_Caps_Lock);
 num_lock = XKeysymToKeycode(dpy, XK_Num_Lock);
 scroll_lock = XKeysymToKeycode(dpy, XK_Scroll_Lock);
 quit_keycode = XKeysymToKeycode(dpy, XK_Q);
 xlock_keycode = XKeysymToKeycode(dpy, XK_L);

 return;
}

void remap(map)
	struct map *map;
{
 if (current_map == map)
	return;

 XChangeKeyboardMapping(dpy, first, key_per_code, map->map, max-first+1);
 PutText(map->text, map->len);
 ChangeIcon(map->pixmap);
 if (fvwm95)
	set_icon_name(map->text);

 current_map = map;

 return;
}

void PutText(text,len)
	char *text;
	int len;
{
 int width,x_pos;

 XClearWindow(dpy, win);
 XDrawRectangle(dpy, win, gc, 0, 0, win_w-1, win_h-1);

 width = XTextWidth(fs, text, len);
 x_pos = (win_w-width) >> 1;
 XDrawString(dpy, win, gc, x_pos, y_pos, text, strlen(text));

 return;
}

Bool GrabKey()
{
 KeySym keysym;

 if (keyname == NULL)
	keyname = GetResource("modifier");
 if (keyname == NULL)
	keyname = "Multi_key";

 keysym = XStringToKeysym(keyname);
 if (keysym == NoSymbol || (keycode = XKeysymToKeycode(dpy, keysym)) == 0) {
	warn("unable to grab the \"%s\" key", keyname);

	return False;
 }

 XGrabKey(dpy, keycode, AnyModifier, DefaultRootWindow(dpy), True,
	      GrabModeAsync, GrabModeAsync);

 return True;
}

Bool RegrabKey()
{
 XEvent event;
 Cursor cursor;
 KeyCode caps_lock, num_lock, scroll_lock;

 if (grabbed)
	XUngrabKey(dpy, keycode, AnyModifier, RootWindow(dpy, scr));

 if (!(cursor = XCreateFontCursor(dpy, XC_exchange)))
	warn("can't create font cursor");
 else
	if (XGrabPointer(dpy, win, False,
			 ButtonPressMask | ButtonReleaseMask,
			 GrabModeSync, GrabModeAsync, RootWindow(dpy,scr),
			 cursor, CurrentTime) != GrabSuccess)
				warn("unable to grab the pointer");

 caps_lock = XKeysymToKeycode(dpy, XK_Caps_Lock);
 num_lock = XKeysymToKeycode(dpy, XK_Num_Lock);
 scroll_lock = XKeysymToKeycode(dpy, XK_Scroll_Lock);

 do 
	do
		XNextEvent(dpy, &event);
	while (event.type != KeyPress);
 while (event.xkey.keycode == caps_lock ||
	event.xkey.keycode == num_lock ||
	event.xkey.keycode == scroll_lock);

 XUngrabPointer(dpy, CurrentTime);

 if (cursor)
	XFreeCursor(dpy, cursor);

 XGrabKey(dpy, event.xkey.keycode, AnyModifier, DefaultRootWindow(dpy),
	  True, GrabModeAsync, GrabModeAsync);

 XBell(dpy, 0);
 keycode = event.xkey.keycode;

 return True;
}

void ChangeIcon(pixmap)
	Pixmap pixmap;
{
 if (have_icons) {
	h.icon_pixmap = pixmap;
	h.flags = IconPixmapHint;

	XSetWMHints(dpy, win, &h);

#ifdef HAVE_XPM
	XSetWindowBackgroundPixmap(dpy, icon_win, pixmap);
	XClearWindow(dpy, icon_win);
#endif

 }

 return;
}

Bool get_x_y(res, X_pos, Y_pos, X_def, Y_def, X, Y)
	char *res;
	char *X_pos, *Y_pos;
	int X_def, Y_def;
	int *X, *Y;
{
 if (X_pos && Y_pos) {
	if (!(*X = atoi(X_pos)) && strcmp(X_pos, "0")) {
		warn("can't convert string \"%s\" to integer", X_pos);
		*X = X_def, *Y = Y_def;

		return False;
	}
	if (!(*Y = atoi(Y_pos)) && strcmp(Y_pos, "0")) {
		warn("can't convert string \"%s\" to integer", Y_pos);
		*X = X_def, *Y = Y_def;

		return False;
	}

	return True;
 }

 if ((X_pos = GetResource(res))) {
	if (sscanf(X_pos, "%d %d", X, Y) != 2) {
		warn("can't convert string \"%s\" to %s property", X_pos, res);
		*X = X_def, *Y = Y_def;

		return False;
	}

	return True;
 }
 *X = X_def, *Y = Y_def;

 return False;
}

void LaunchXLock()
{
 int save_index;
 Bool save_popup;
 struct map *ptr;

 if ((save_index = current_map->index) != STANDARD_MAP) {
	for (ptr = current_map; ptr->index != STANDARD_MAP; ptr = ptr->next);
	remap(ptr);
 }

 save_popup = popup;
 popup = False;

 if (system(xlock_string) < 0)
	warn("can't launch \"%s\"", xlock_string);

 popup = save_popup;

 if (save_index != STANDARD_MAP) {
	for (ptr = current_map; ptr->index != save_index; ptr = ptr->next);
	remap(ptr);
 }

 return;
}
