#ifdef HAVE_STRINGS_H
#include <strings.h>
#endif
#include <stdlib.h>

#include <X11/Intrinsic.h>
#include <X11/Xatom.h>
#include "global.h"
#include "map.h"
#include "version.h"

#ifdef HAVE_STRRCHR
#define rindex(s,c) strrchr(s,c)
#endif

#define BG_COLOR 0
#define FG_COLOR 1

#define XES_ATOM	"XES_ALREADY_RUNNING"

PUBLIC(void, SetWMHints, (Display *dpy,
			  char **argv,
			  Cardinal argc,
			  Window w,
			  int x,
			  int y,
			  int width,
			  int height,
			  int min_w,
			  int min_h,
			  int max_w,
			  int max_h));
PUBLIC(int, decoration_delete,(void));
PUBLIC(int, decoration_delete_olwm,(void));
PUBLIC(int, decoration_delete_mwm,(void));
PUBLIC(void, ParseCommandLine, (Cardinal *argc, char **argv));
PUBLIC(void, InitColors, (void));
PUBLIC(void, InitFont, (void));
PUBLIC(void, InitGC, (void));
PUBLIC(void, Quit, (int status));
PUBLIC(void, QuitIfAlreadyRunning, (void));

EXTERN(char*, GetResource, (char *name));
EXTERN(Bool, GetBoolResource, (char *name));
EXTERN(void, remap, (struct map *map));
EXTERN(void, store_maps, (void));
EXTERN(Bool, create_icons, (void));
EXTERN(Bool, GrabKey, (void));
EXTERN(void, InitKeycodes, (void));
EXTERN(void, MainLoop, (void));
EXTERN(void, intercept_signals, (void));
EXTERN(void, InitializeRDB, (void));
EXTERN(void, set_icon_name, (char *name));
EXTERN(Bool, get_x_y, (char *res, char *X_pos, char *Y_pos,
		       int X_def, int Y_def, int *X, int *Y));


EXTERN(int, IOHandler, (Display *display));

Display *dpy;
Atom xes_atom;
char *dpyname = NULL, *app_name;
char *X_pos = NULL, *Y_pos = NULL, *XI_pos = NULL, *YI_pos = NULL;
int scr, win_w, win_h, X, Y, XI, YI, y_pos;
Colormap cmap;
Window win = (Window)0;
int icon_size = 0;

#ifdef HAVE_XPM
Window icon_win;
#endif

XWMHints h;
GC gc;
unsigned long ncolor[2];
Font font;
XFontStruct *fs;
Bool have_icons = False, grabbed = False, popup = False, iconic = False,
     nokoi = False, noiso = False, noalt = False, nowin = False,
     ignore_running = False, fvwm95 = False;
char *bg = NULL, *fg = NULL, *fontname = NULL, *keyname = NULL, *wmname = NULL;
char *xlock_string;

extern KeyCode keycode;
extern struct map *current_map;

#ifndef HAVE_XPM
extern Pixmap icon_mask;
#endif

Atom protocol[2], wmproto;

void usage(status)
	int status;
{

#define usg(string)	(void)fputs(string,stderr)

 usg("Usage:\n        ");
 usg(app_name);
 usg(" [-options ...]\n\n");
 usg("where options include:\n");
 usg("    -usage | -help | -?             print this message\n");
 usg("    -display <displayname>          X server to contact\n");
 usg("    -bg <color>                     background color\n");
 usg("    -fg <color>                     foreground color\n");
 usg("    -fn <fontname>                  normal text font\n");
 usg("    -wpos <x> <y>                   set the initial position of program window\n");
 usg("    -modifier <keyname>             key-modifier name\n");
 usg("    -iconic                         start as an icon\n");
 usg("    -48                             set icon size to 48x48 pixels\n");
 usg("    -64                             set icon size to 64x64 pixels\n");
 usg("    -ipos <x> <y>                   set the initial position of icon window\n");
 usg("    -wm <name>                      window manager name\n");
 usg("    -popup                          program window always on top\n");
 usg("    -nokoi                          disable KOI8-R encoding\n");
 usg("    -noiso                          disable ISO8859-5 encoding\n");
 usg("    -noalt                          disable ALT encoding\n");
 usg("    -nowin                          disable WIN encoding\n");
 usg("    -anyway                         ignore if already running\n");
 usg("    -v | -version                   print version information\n");

#undef usg

 exit(status);
}

void main(argc, argv)
	Cardinal argc;
	char **argv;
{
 char *pos, *is;

 if ((pos = rindex(argv[0],(int)'/')) == NULL)
	app_name = argv[0];
 else
	app_name = ++pos;

 ParseCommandLine(&argc, argv);

 if ((dpy = XOpenDisplay(dpyname)) == NULL) {
	warn("unable to open display \"%s\"", XDisplayName(dpyname));
	exit(1);
 }
 QuitIfAlreadyRunning();

 scr = DefaultScreen(dpy);
 cmap = DefaultColormap(dpy, scr);

 intercept_signals();

 wmproto = XInternAtom(dpy, "WM_PROTOCOLS", True);
 protocol[0] = XInternAtom(dpy, "WM_DELETE_WINDOW", True);
 protocol[1] = XInternAtom(dpy, "WM_SAVE_YOURSELF", True);

 InitializeRDB();

 if (!nokoi)
	nokoi = GetBoolResource("nokoi");
 if (!noiso)
	noiso = GetBoolResource("noiso");
 if (!noalt)
	noalt = GetBoolResource("noalt");
 if (!nowin)
	nowin = GetBoolResource("nowin");

 if (!popup)
	popup = GetBoolResource("popup");

 store_maps();

 if (!(xlock_string = GetResource("xlockString")))
	xlock_string = "xlock -mode random";

 if ((pos = GetResource("standardlabel"))) {
	current_map->text = pos;
	current_map->len = strlen(pos);
 }

 if (!nokoi)
	if ((pos = GetResource("koilabel"))) {
		current_map->next->text = pos;
		current_map->next->len = strlen(pos);
	}

 if (!noiso)
	if ((pos = GetResource("isolabel"))) {
		current_map->next->next->text = pos;
		current_map->next->next->len = strlen(pos);
	 }

 if (!noalt)
	if ((pos = GetResource("altlabel"))) {
		current_map->prev->prev->text = pos;
		current_map->prev->prev->len = strlen(pos);
	}

 if (!nowin)
	if ((pos = GetResource("winlabel"))) {
		current_map->prev->text = pos;
		current_map->prev->len = strlen(pos);
	}

 InitFont();
 InitColors();

 (void) get_x_y("windowPosition", X_pos, Y_pos, 100, 0, &X, &Y);

 win = XCreateSimpleWindow(dpy, RootWindow(dpy, scr), X, Y, win_w, win_h, 5,
			   BlackPixel(dpy, scr), ncolor[BG_COLOR]);
 grabbed = GrabKey();

 if ((is = GetResource("iconSize")))
	switch (icon_size = atoi(is)) {
		case 0:
			warn("can't convert string \"%s\" to iconSize property",
			     is);
		break;
		case 48:
		case 64:
		break;
		default:
			warn("iconSize property must be 48 or 64");
		break;
	}
 have_icons = create_icons();

 SetWMHints(dpy, argv, argc, win, 100, 0, win_w, win_h,
	    win_w, win_h, win_w, win_h);
 XSetWMProtocols(dpy, win, protocol, 2);

 XSetIOErrorHandler(IOHandler);

 InitGC();
 InitKeycodes();

 XSelectInput(dpy, win, KeyPressMask | ButtonPressMask | ExposureMask |
	      VisibilityChangeMask);

#ifdef HAVE_XPM
 XSelectInput(dpy, icon_win, KeyPressMask);
#endif

 XMapWindow(dpy,win);

 MainLoop();

 Quit(0);
}

void SetWMHints(dpy, argv, argc, w, x, y, width, height,
		min_w, min_h, max_w, max_h)
	Display *dpy;
	char **argv;
	Cardinal argc;
	Window w;
	int x, y, width, height, min_w, min_h, max_w, max_h;
{
 XSizeHints s;
 XClassHint c;
 XTextProperty win_name, icon_name;
 char *class;

 class = app_class;

 s.flags = PPosition | PSize | PMinSize | PMaxSize;
 s.min_width = min_w;
 s.min_height = min_h;
 s.max_width = max_w;
 s.max_height = max_h;

 if (!iconic)
	iconic = GetBoolResource("iconic");

 h.initial_state = iconic ? IconicState : NormalState;
 h.input = True;
 if (have_icons) {
#ifdef HAVE_XPM
	h.icon_window = icon_win;
#endif
	h.icon_pixmap = current_map->pixmap;
#ifndef HAVE_XPM
	h.icon_mask = icon_mask;
	h.flags = StateHint | InputHint | IconPixmapHint | IconMaskHint;
#else
	h.flags = StateHint | InputHint | IconWindowHint | IconPixmapHint;
#endif

	if (get_x_y("iconPosition", XI_pos, YI_pos, 0, 0, &XI, &YI)) {
		h.icon_x = XI, h.icon_y = YI;
		h.flags |= IconPositionHint;
	}
 } else
	h.flags = StateHint | InputHint;

 c.res_name = app_name;
 c.res_class = app_class;

 switch (decoration_delete()) {
	case 0 :
		XSetWMProperties(dpy, w, NULL, NULL, argv, argc, &s, &h, &c);
	break;
	case 1 :
		if (!XStringListToTextProperty(&app_name, 1, &icon_name)) {
			warn("Out of memory");
			Quit(1);
		}
		XSetWMProperties(dpy, w, NULL, &icon_name,
				 argv, argc, &s, &h, &c);
	break;
	case 2 :
		if (!XStringListToTextProperty(&class, 1, &win_name)) {
			warn("Out of memory");
			Quit(1);
		}
		XSetWMProperties(dpy, w, &win_name, NULL,
				 argv, argc, &s, &h, &c);
	break;
	case 3 :
		if (!XStringListToTextProperty(&class,1,&win_name) ||
		    !XStringListToTextProperty(&app_name,1,&icon_name)) {
			warn("Out of memory");
			Quit(1);
		}
		XSetWMProperties(dpy, w, &win_name, &icon_name,
				 argv, argc, &s, &h, &c);
	break;
 }

#ifdef HAVE_XPM
 if (have_icons) {
	XSetWindowBackgroundPixmap(dpy, icon_win, current_map->pixmap);
	XClearWindow(dpy, icon_win);
 }
#endif

 return;
}

int decoration_delete()
{
 if (!wmname)
	wmname = GetResource("windowmanager");

 if (!wmname)
	return 3;

 if (!strcasecmp(wmname,"olwm") || !strcasecmp(wmname,"olvwm"))
	return decoration_delete_olwm();

 if (!strcasecmp(wmname,"mwm") || !strcasecmp(wmname,"mvwm") ||
     !strcasecmp(wmname,"dtwm"))
	return decoration_delete_mwm();

 if (!strncasecmp(wmname,"fvwm95", 6)) {
	fvwm95 = True;
	set_icon_name(current_map->text);
	return 1;
 }

 return 3;
}

int decoration_delete_olwm()
{
 Atom decor_del, header, icon;

 if ((decor_del = XInternAtom(dpy, "_OL_DECOR_DEL", True)) == None) {
	warn("can't intern atom \"_OL_DECOR_DEL\"");
	return 3;
 }

 if ((header = XInternAtom(dpy, "_OL_DECOR_HEADER", True)) == None) {
	warn("can't intern atom \"_OL_DECOR_HEADER\"");
	if ((icon = XInternAtom(dpy, "_OL_DECOR_ICON_NAME", True)) == None) {
		warn("can't intern atom \"_OL_DECOR_ICON_NAME\"");
		return 3;
	} else {
		if (!have_icons)
			return 3;
		XChangeProperty(dpy, win, decor_del, XA_ATOM, 32,
				PropModePrepend, (unsigned char*) &icon, 1);                                     
		return 2;
	}
 }

 XChangeProperty(dpy, win, decor_del, XA_ATOM, 32, PropModePrepend,
		 (unsigned char*) &header, 1);                                     
 
 if (!have_icons)
	return 1;

 if ((icon = XInternAtom(dpy, "_OL_DECOR_ICON_NAME", True)) == None) {
	warn("can't intern atom \"_OL_DECOR_ICON_NAME\"");
	return 1;
 }

 XChangeProperty(dpy, win, decor_del, XA_ATOM, 32, PropModePrepend,
		 (unsigned char*) &icon, 1);                                     

 return 0;
}

int decoration_delete_mwm()
{
 Atom hints;

 struct {
	long flags;
	long functions;
	long decorations;
	unsigned long input_mode;
	long status;
 } data;

 if(!(hints = XInternAtom(dpy, "_MOTIF_WM_HINTS", True))) {
	warn("can't intern atom \"_MOTIF_WM_HINTS\"");
	return 3;
 }

#ifdef HAVE_BZERO
 bzero(&data, sizeof(data));
#else
 memset(&data, 0, sizeof(data));
#endif

 data.flags = 2;
 data.decorations = 2;

 XChangeProperty(dpy, win, hints, hints, 32, PropModeReplace,
		 (unsigned char*) &data, sizeof(data)/4);

 return 1;
}
 
void ParseCommandLine(argc, argv)
	Cardinal *argc;
	char **argv;
{
 for (argv++; *argv != NULL; argv++) {
	if (!strcmp(*argv, "-usage") || !strcmp(*argv, "-help") ||
	    !strcmp(*argv, "-?"))
		usage(0);
	else
	if (!strcmp(*argv, "-display"))
		if (*(++argv) != NULL)
			dpyname = *argv;
		else
			usage(1);
	else
	if (!strcmp(*argv, "-bg"))
		if (*(++argv) != NULL)
			bg = *argv;
		else
			usage(1);
	else
	if (!strcmp(*argv, "-fg"))
		if (*(++argv) != NULL)
			fg = *argv;
		else
			usage(1);
	else
	if (!strcmp(*argv, "-fn"))
		if (*(++argv) != NULL)
			fontname = *argv;
		else
			usage(1);
	else
	if (!strcmp(*argv, "-modifier"))
		if (*(++argv) != NULL)
			keyname = *argv;
		else
			usage(1);
	else
	if (!strcmp(*argv, "-wm"))
		if (*(++argv) != NULL)
			wmname = *argv;
		else
			usage(1);
	else
	if (!strcmp(*argv, "-wpos"))
		if (*(++argv) != NULL) {
			X_pos = *argv;
			if (*(++argv) != NULL)
				Y_pos = *argv;
			else
				usage(1);
		} else
			usage(1);
	else
	if (!strcmp(*argv, "-ipos"))
		if (*(++argv) != NULL) {
			XI_pos = *argv;
			if (*(++argv) != NULL)
				YI_pos = *argv;
			else
				usage(1);
		} else
			usage(1);
	else
	if (!strcmp(*argv, "-iconic"))
		iconic = True;
	else
	if (!strcmp(*argv, "-48"))
		icon_size = 48;
	else
	if (!strcmp(*argv, "-64"))
		icon_size = 64;
	else
	if (!strcmp(*argv, "-popup"))
		popup = True;
	else
	if (!strcmp(*argv, "-nokoi"))
		nokoi = True;
	else
	if (!strcmp(*argv, "-noiso"))
		noiso = True;
	else
	if (!strcmp(*argv, "-noalt"))
		noalt = True;
	else
	if (!strcmp(*argv, "-nowin"))
		nowin = True;
	else
	if (!strcmp(*argv, "-anyway"))
		ignore_running = True;
	else
	if (!strcmp(*argv, "-v") || !strcmp(*argv, "-version")) {
		fputs(version, stderr);
		exit(0);
	}
	else
		usage(1);
 }

 *argc = 0;

 return;
}

void InitColors()
{
 Visual *v;
 XColor bg_color, fg_color;
 char *bg_name, *fg_name;

 v = DefaultVisual(dpy, scr);
 if (v->class != PseudoColor && v->class != TrueColor &&
     v->class != DirectColor) {
	ncolor[BG_COLOR] = WhitePixel(dpy, scr);
	ncolor[FG_COLOR] = BlackPixel(dpy, scr);

	return;
 }

 if ((bg_name = (bg == NULL) ? GetResource("background") : bg) == NULL)
	bg_name = "white";
 if ((fg_name = (fg == NULL) ? GetResource("foreground") : fg) == NULL)
	fg_name = "black";

 if (!XParseColor(dpy, cmap, bg_name, &bg_color))
	warn("invalid color name \"%s\". Using monochrome", bg_name);
 else
	if (!XParseColor(dpy, cmap, fg_name, &fg_color))
		warn("invalid color name \"%s\". Using monochrome", fg_name);
	else
		if (XAllocColor(dpy, cmap, &bg_color) &&
		    XAllocColor(dpy, cmap, &fg_color)) {
			ncolor[BG_COLOR] = bg_color.pixel;
			ncolor[FG_COLOR] = fg_color.pixel;

			return;
		} else 
			warn("can't allocate colormap entry. Using monochrome");

 ncolor[BG_COLOR] = WhitePixel(dpy,scr);
 ncolor[FG_COLOR] = BlackPixel(dpy,scr);

 return;
}

void InitFont()
{
 char *f;
 int len;
 struct map *ptr;

 if (fontname != NULL)
	f = fontname;
 else
	if ((f = GetResource("font")) == NULL)
		f = "fixed";
 font = XLoadFont(dpy, f);

 fs = XQueryFont(dpy, font);

 len = XTextWidth(fs, current_map->text, current_map->len);
 for (ptr = current_map->next; ptr->index != STANDARD_MAP; ptr = ptr->next)
	if (ptr->enabled) {
		int l;

		l = XTextWidth(fs, ptr->text, ptr->len);
		if (l > len)
			len = l;
	}

 win_h = fs->ascent + fs->descent + 12;
 win_w = len + 12;
 y_pos = fs->ascent + 7;

 return;
}

void InitGC()
{
 XGCValues val;
 unsigned long mask;

 val.function = GXcopy;
 val.background = ncolor[BG_COLOR];
 val.foreground = ncolor[FG_COLOR];
 val.font = font;

 mask = GCFunction | GCForeground | GCBackground | GCFont;

 gc = XCreateGC(dpy, win, mask, &val);

 return;
}

void Quit(status)
	int status;
{
 struct map *ptr;

 XDeleteProperty(dpy, DefaultRootWindow(dpy), xes_atom);

 for (ptr = current_map; ptr->index != STANDARD_MAP; ptr = ptr->next);
 if (ptr->map)
	remap(ptr);

 if (status != 255) {
	if (grabbed)
		XUngrabKey(dpy, keycode, AnyModifier, DefaultRootWindow(dpy));
	if (win)
		XDestroyWindow(dpy, win);
 }
 XCloseDisplay(dpy);

 exit(status);
}

void QuitIfAlreadyRunning()
{
 XTextProperty prop;
 char *s;

 if ((xes_atom = XInternAtom(dpy, XES_ATOM, False)) == None) {
	warn("may be already running. Can't check");

	return;
 }

 if (XGetTextProperty(dpy, DefaultRootWindow(dpy), &prop, xes_atom)) {
	if (ignore_running)
		return;

	warn("already running on display \"%s\"", XDisplayName(dpyname));
	XCloseDisplay(dpy);

	exit(1);
 }

 s = XES_ATOM;
 if (!XStringListToTextProperty(&s, 1, &prop)) {
	warn("can't set text property to the root window");

	return;
 }

 XSetTextProperty(dpy, DefaultRootWindow(dpy), &prop,xes_atom);

 return;
}
