#ifdef HAVE_STRINGS_H
#include <strings.h>
#else
#include <stdlib.h>
#endif

#include <X11/Intrinsic.h>
#include <stdlib.h>
#include "global.h"

#ifndef MAXPATHLEN
#define MAXPATHLEN	1024
#endif

#ifndef MAXNAMELEN
#define MAXNAMELEN	256
#endif

struct {
	char name;
	char *subst;
} subs[] = {
		{ 'N', "Xes" },
		{ 'T', "app-defaults" },
		{ 'L', NULL },
		{ '\0', NULL }
  };

PUBLIC(Bool, ResolvePath, (char **ptr,char *path));
PUBLIC(void, InitializeRDB, (void));
PUBLIC(char*, GetResource, (char *name));
PUBLIC(Bool, GetBoolResource, (char *name));

static XrmDatabase DB;

extern Display *dpy;
extern char *app_name;

Bool ResolvePath(ptr, path)
	char **ptr, *path;
{
 int i;

#ifdef HAVE_BZERO
 bzero(path, MAXPATHLEN);
#else
 memset(path, 0, MAXPATHLEN);
#endif

 while (**ptr != '\0' && **ptr != ':') {
	if (**ptr == '%') {
		(*ptr)++;
		if (!(**ptr))
			return False;
		for (i=0; subs[i].name; i++)
			if ((**ptr == subs[i].name) && subs[i].subst) {
				(void) strcat(path, subs[i].subst);
				path += strlen(subs[i].subst);
				break;
			}
	} else {
		*path = **ptr;
		path++;
	}
	(*ptr)++;
 }

 return True;
}

void InitializeRDB()
{
 char path[MAXPATHLEN];
 char *env, *ptr;

 XrmInitialize();

 if ((env = getenv("HOME")) == NULL)
	env = ".";
 sprintf(path, "%s/%s", env, app_class);

 DB = XrmGetFileDatabase(path);

 if ((ptr = getenv("XFILESEARCHPATH")) == NULL)
	ptr = XFILESEARCHPATHDEFAULT;

 if ((subs[2].subst = getenv("LANG")) != NULL)
	if (!strcmp(subs[2].subst, "C"))
		subs[2].subst = NULL;

 while (ResolvePath(&ptr, path)) {
	if (path[0] == '\0')
		break;

	XrmCombineFileDatabase(path, &DB, False);
	if (*ptr)
		ptr++;
 }

 return;
}

char *GetResource(name)
	char *name;
{
 char *value;

 value = XGetDefault(dpy, app_name, name);

 if (value == NULL && strcmp(app_name, app_class))
	value = XGetDefault(dpy, app_class, name);

 if (value == NULL && DB != NULL) {
	char xrm_name[MAXNAMELEN];
	char xrm_class[MAXNAMELEN];
	char *type;
	XrmValue val;

	sprintf(xrm_name, "%s.%s", app_name, name);
	sprintf(xrm_class, "%s.%s",app_class, name);

	if (XrmGetResource(DB, xrm_name, xrm_class, &type, &val))
		value = (char*) val.addr;
 }

 return value;
}

Bool GetBoolResource(name)
	char *name;
{
 char *value;

 if ((value = GetResource(name)) == NULL)
	return False;

 if (!strcasecmp(value, "true") || !strcasecmp(value, "on"))
	return True;

 return False;
}
